{
    "cells": [
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": [
                "# Klassen und Objekte\n",
                "\n",
                "[⬇️ Jupyter Notebook herunterladen](../code/python-klassen.ipynb){download=\"python-klassen.ipynb\"}\n",
                "\n",
                "\n",
                "Die objektorientierte Programmierung (OOP) ist ein wichtiges Konzept in der Softwareentwicklung. Es hilft besonders dabei, große Codebasen für komplexe Data-Science-Projekte zu organisieren und zu verwalten. \n",
                "\n",
                "In diesem Abschnitt lernen wir:\n",
                "\n",
                "- Wie man Klassen und Objekte erstellt\n",
                "- Wie man Methoden und Attribute hinzufügt\n",
                "- Wie man Vererbung verwendet, um die Funktionalität zu erweitern\n",
                "\n",
                "\n",
                "## Einführung \n",
                "\n",
                "Eine **Klasse** ist wie ein Bauplan oder eine Vorlage. Stellen Sie sich eine Klasse wie die Blaupause eines Hauses vor. Eine Blaupause beschreibt, wie das Haus aussehen soll – welche Räume es hat, wie groß sie sind, und so weiter. In ähnlicher Weise beschreibt eine Klasse, welche Daten (Eigenschaften) und Funktionen (Methoden) ein bestimmtes Objekt haben wird.\n",
                "\n",
                "Ein **Objekt** ist eine konkrete Instanz einer Klasse. Wenn wir bei unserem Beispiel des Hauses bleiben, wäre ein Objekt ein tatsächliches Haus, das nach der Blaupause gebaut wurde. Jedes Haus, das nach derselben Blaupause gebaut wird, hat dieselben grundlegenden Eigenschaften, kann aber unterschiedliche Farben, Möbel oder Bewohner haben. In der Programmierung ist ein Objekt also ein spezifisches Exemplar einer Klasse mit eigenen, spezifischen Daten.\n",
                "\n",
                "Wozu sind Klassen und Objekte nützlich:\n",
                "\n",
                "- **Organisation**: Sie helfen, den Code in überschaubare, wiederverwendbare Teile zu zerlegen.\n",
                "- **Wiederverwendbarkeit**: Einmal definierte Klassen können überall im Programm verwendet werden, um konsistente Objekte zu erstellen.\n",
                "- **Erweiterbarkeit**: Neue Funktionen und Eigenschaften können leicht zu bestehenden Klassen hinzugefügt werden, ohne den gesamten Code neu schreiben zu müssen.\n",
                "- **Lesbarkeit**: Der Code wird klarer und einfacher zu verstehen, weil die Struktur und das Verhalten der Daten klar definiert sind.\n",
                "\n",
                "\n",
                "Betrachten wir eine einfache Klasse und ein Objekt, um das Konzept klarer zu machen:"
            ]
        },
        {
            "cell_type": "code",
            "metadata": {},
            "source": [
                "# Definiere eine einfache Klasse, die eine Person repräsentiert\n",
                "class Person:\n",
                "    \"\"\"Eine Klasse, die eine Person mit Namen und Alter repräsentiert.\"\"\"\n",
                "\n",
                "    def __init__(self, name, age):\n",
                "        \"\"\"Initialisiere die Person mit Name und Alter.\"\"\"\n",
                "        self.name = name\n",
                "        self.age = age\n",
                "\n",
                "    def greet(self):\n",
                "        \"\"\"Gibt eine Begrüßung zurück.\"\"\"\n",
                "        return f\"Hallo, mein Name ist {self.name} und ich bin {self.age} Jahre alt.\"\n",
                "\n",
                "# Erstelle ein Objekt der Klasse Person\n",
                "person1 = Person(\"Alice\", 30)\n",
                "\n",
                "# Verwende die Methode des Objekts\n",
                "print(person1.greet())"
            ],
            "execution_count": null,
            "outputs": []
        },
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": [
                "In diesem Beispiel:\n",
                "\n",
                "1. **Klasse `Person`**: Dies ist die Blaupause. Sie beschreibt, dass jede Person einen Namen und ein Alter hat und sich vorstellen kann.\n",
                "2. **Objekt `person1`**: Dies ist ein spezifisches Exemplar der Klasse `Person`. Es repräsentiert eine konkrete Person namens Alice, die 30 Jahre alt ist.\n",
                "3. **Docstring**: Die Kommentare in dreifachen Anführungszeichen direkt unter der Klassendefinition und den Methoden beschreiben, was die Klasse und die Methoden machen.\n",
                "4. **Methoden und Attribute**: Das Objekt `person1` verwendet die Methode `greet()`, um eine Begrüßung auszugeben, die auf den Attributen `name` und `age` basiert.\n",
                "\n",
                "\n",
                "\n",
                "## Erstellung einer Klasse\n",
                "\n",
                "In diesem Beispiel behandeln wir ausführlich die Definition einer Klasse `DataPoint`, die eine einzelne Beobachtung in einem Datensatz darstellt. "
            ]
        },
        {
            "cell_type": "code",
            "metadata": {},
            "source": [
                "# Definiere eine einfache Klasse, die eine Datenbeobachtung repräsentiert\n",
                "class DataPoint:\n",
                "    \"\"\"Eine Klasse, die eine Datenbeobachtung mit Merkmalen und Label repräsentiert.\"\"\"\n",
                "\n",
                "    def __init__(self, features, label):\n",
                "        \"\"\"Initialisiere den Datenpunkt mit Merkmalen und einem Label.\"\"\"\n",
                "        self.features = features\n",
                "        self.label = label\n",
                "\n",
                "    def describe(self):\n",
                "        \"\"\"Gibt einen formatierten String zurück, der den Datenpunkt beschreibt.\"\"\"\n",
                "        return f\"Merkmale: {self.features}, Label: {self.label}\"\n",
                "\n",
                "# Erstelle eine Instanz (Objekt) der DataPoint-Klasse\n",
                "sample_point = DataPoint([0.25, 0.75, 0.5], 'Positiv')\n",
                "\n",
                "# Zeige die Beschreibung des Datenpunkts an\n",
                "print(sample_point.describe())"
            ],
            "execution_count": null,
            "outputs": []
        },
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": [
                "**Definition der Klasse `DataPoint`**:\n",
                "\n",
                "1. **Klassenkopf**: \n",
                "   ```python\n",
                "   class DataPoint:\n",
                "   ```\n",
                "   Dies definiert eine neue Klasse namens `DataPoint`. \n",
                "\n",
                "2. **Dokumentations-String**:\n",
                "   ```python\n",
                "   \"\"\"Eine Klasse, die eine Datenbeobachtung mit Merkmalen und Label repräsentiert.\"\"\"\n",
                "   ```\n",
                "   Dies ist eine kurze Beschreibung, was die Klasse macht. Es hilft anderen schnell zu verstehen, wofür diese Klasse gedacht ist.\n",
                "\n",
                "**Initialisierungsmethode `__init__`**\n",
                "\n",
                "3. **Konstruktor-Methode**:\n",
                "   ```python\n",
                "   def __init__(self, features, label):\n",
                "       \"\"\"Initialisiere den Datenpunkt mit Merkmalen und einem Label.\"\"\"\n",
                "       self.features = features\n",
                "       self.label = label\n",
                "   ```\n",
                "\n",
                "   Die `__init__`-Methode wird automatisch aufgerufen, wenn eine neue Instanz der Klasse erstellt wird. Sie initialisiert die Eigenschaften (Attribute) des Objekts. In diesem Fall nimmt sie zwei Argumente `features` und `label` und speichert sie in den Eigenschaften `self.features` und `self.label` des Objekts.\n",
                "\n",
                "   - `self`: Bezieht sich auf das aktuelle Objekt der Klasse und wird verwendet, um auf die Attribute und Methoden des Objekts zuzugreifen.\n",
                "   - `features`: Ein Satz von Merkmalen, die den Datenpunkt beschreiben.\n",
                "   - `label`: Ein Etikett, das dem Datenpunkt zugewiesen ist, z.B. 'Positiv' oder 'Negativ'.\n",
                "\n",
                "**Methode `describe`**\n",
                "\n",
                "4. **Beschreibungsmethode**:\n",
                "   ```python\n",
                "   def describe(self):\n",
                "       \"\"\"Gibt einen formatierten String zurück, der den Datenpunkt beschreibt.\"\"\"\n",
                "       return f\"Merkmale: {self.features}, Label: {self.label}\"\n",
                "   ```\n",
                "   Diese Methode gibt eine lesbare Beschreibung des Datenpunkts zurück, die die Merkmale und das Label des Datenpunkts enthält.\n",
                "\n",
                "**Erstellung einer Instanz und Nutzung der Methode**\n",
                "\n",
                "5. **Erstellung einer Instanz**:\n",
                "   ```python\n",
                "   sample_point = DataPoint([0.25, 0.75, 0.5], 'Positiv')\n",
                "   ```\n",
                "   Hier erstellen wir ein Objekt namens `sample_point` von der Klasse `DataPoint`. Die Merkmale werden als Liste `[0.25, 0.75, 0.5]` und das Label als `'Positiv'` übergeben.\n",
                "\n",
                "6. **Aufruf der Methode `describe`**:\n",
                "   ```python\n",
                "   print(sample_point.describe())\n",
                "   ```\n",
                "   Diese Zeile ruft die `describe`-Methode des `sample_point`-Objekts auf und druckt die zurückgegebene Beschreibung des Datenpunkts aus.\n",
                "\n",
                "\n",
                "In unserem Beispiel haben wir also eine `DataPoint`-Klasse erstellt, die Merkmale und ein Label enthält, und eine Methode, um diese Daten in einem lesbaren Format anzuzeigen.\n",
                "\n",
                "## Methoden und Attribute\n",
                "\n",
                "Das Hinzufügen weiterer Funktionalität zu einer Klasse macht sie vielseitiger. Lassen Sie uns unsere Klasse erweitern, um eine Sammlung von Datenpunkten zu verwalten:"
            ]
        },
        {
            "cell_type": "code",
            "metadata": {},
            "source": [
                "# Definiere eine Klasse zur Verwaltung eines Datensatzes\n",
                "class DataSet:\n",
                "    \"\"\"Eine Klasse, die eine Sammlung von Datenpunkten repräsentiert.\"\"\"\n",
                "\n",
                "    def __init__(self):\n",
                "        \"\"\"Initialisiere die leere Sammlung von Datenpunkten.\"\"\"\n",
                "        self.data_points = []\n",
                "\n",
                "    def add_data_point(self, data_point):\n",
                "        \"\"\"Füge einen Datenpunkt zur Sammlung hinzu.\"\"\"\n",
                "        self.data_points.append(data_point)\n",
                "\n",
                "    def display_data(self):\n",
                "        \"\"\"Zeige alle Datenpunkte in der Sammlung an.\"\"\"\n",
                "        for data_point in self.data_points:\n",
                "            print(data_point.describe())\n",
                "\n",
                "# Erstelle eine Instanz der DataSet-Klasse\n",
                "my_data_set = DataSet()\n",
                "\n",
                "# Füge mehrere DataPoint-Objekte zum Datensatz hinzu\n",
                "my_data_set.add_data_point(DataPoint([0.1, 0.2, 0.5], 'Negativ'))\n",
                "my_data_set.add_data_point(DataPoint([0.8, 0.4, 0.3], 'Positiv'))\n",
                "\n",
                "# Zeige alle Datenpunkte im Datensatz an\n",
                "my_data_set.display_data()"
            ],
            "execution_count": null,
            "outputs": []
        },
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": [
                "**Erklärung**:\n",
                "\n",
                "- Die `DataSet`-Klasse speichert mehrere `DataPoint`-Objekte.\n",
                "- Methoden wie `add_data_point` und `display_data` ermöglichen das Hinzufügen von Datenpunkten und das Anzeigen aller Punkte.\n",
                "\n",
                "## Vererbung\n",
                "\n",
                "Vererbung ermöglicht es uns, bestehende Klassen zu erweitern, um spezialisiertere Versionen zu erstellen, ohne den Code zu wiederholen. Beispielsweise definieren wir einen `TimeSeriesDataPoint`, der `DataPoint` erweitert, indem ein Zeitstempel hinzugefügt wird:"
            ]
        },
        {
            "cell_type": "code",
            "metadata": {},
            "source": [
                "# Definiere eine Klasse, die einen Zeitreihendatenpunkt repräsentiert\n",
                "class TimeSeriesDataPoint(DataPoint):\n",
                "    \"\"\"Eine Klasse, die DataPoint um einen Zeitstempel für Zeitreihendaten erweitert.\"\"\"\n",
                "\n",
                "    def __init__(self, features, label, timestamp):\n",
                "        \"\"\"Initialisiere die Attribute des TimeSeriesDataPoint.\"\"\"\n",
                "        super().__init__(features, label)\n",
                "        self.timestamp = timestamp\n",
                "\n",
                "    def describe(self):\n",
                "        \"\"\"Gibt einen formatierten String zurück, der den Zeitreihendatenpunkt beschreibt.\"\"\"\n",
                "        base_description = super().describe()\n",
                "        return f\"{base_description}, Zeitstempel: {self.timestamp}\"\n",
                "\n",
                "# Erstelle eine Instanz der TimeSeriesDataPoint-Klasse\n",
                "time_series_point = TimeSeriesDataPoint([0.9, 0.1, 0.2], 'Negativ', '2024-09-05T12:00:00')\n",
                "\n",
                "# Zeige die Beschreibung des Zeitreihendatenpunkts an\n",
                "print(time_series_point.describe())"
            ],
            "execution_count": null,
            "outputs": []
        }
    ],
    "metadata": {
        "kernelspec": {
            "name": "python3",
            "language": "python",
            "display_name": "Python 3 (ipykernel)"
        }
    },
    "nbformat": 4,
    "nbformat_minor": 4
}